package sun.nio.ch;

import java.nio.channels.InterruptedByTimeoutException;
import java.nio.BufferOverflowException;
import java.nio.channels.AsynchronousCloseException;
import java.net.InetAddress;
import java.util.concurrent.TimeUnit;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.AlreadyConnectedException;
import java.nio.channels.ConnectionPendingException;
import java.util.concurrent.Future;
import java.nio.channels.CompletionHandler;
import java.security.PrivilegedActionException;
import java.security.AccessController;
import java.security.PrivilegedExceptionAction;
import java.net.SocketAddress;
import java.net.InetSocketAddress;
import java.io.IOException;
import java.nio.channels.ShutdownChannelGroupException;
import sun.misc.Unsafe;

class WindowsAsynchronousSocketChannelImpl extends AsynchronousSocketChannelImpl implements Iocp.OverlappedChannel
{
    private static final Unsafe unsafe;
    private static int addressSize;
    private static final int SIZEOF_WSABUF;
    private static final int OFFSETOF_LEN = 0;
    private static final int OFFSETOF_BUF;
    private static final int MAX_WSABUF = 16;
    private static final int SIZEOF_WSABUFARRAY;
    final long handle;
    private final Iocp iocp;
    private final int completionKey;
    private final PendingIoCache ioCache;
    private final long readBufferArray;
    private final long writeBufferArray;
    
    private static int dependsArch(int n, int n2) {
        return (WindowsAsynchronousSocketChannelImpl.addressSize == 4) ? n : n2;
    }
    
    WindowsAsynchronousSocketChannelImpl(Iocp iocp, boolean b) throws IOException {
        super(iocp);
        long handle = (long)IOUtil.fdVal(this.fd);
        int associate = 0;
        try {
            associate = iocp.associate(this, handle);
        }
        catch (ShutdownChannelGroupException ex) {
            if (b) {
                closesocket0(handle);
                throw ex;
            }
        }
        catch (IOException ex2) {
            closesocket0(handle);
            throw ex2;
        }
        this.handle = handle;
        this.iocp = iocp;
        this.completionKey = associate;
        this.ioCache = new PendingIoCache();
        this.readBufferArray = WindowsAsynchronousSocketChannelImpl.unsafe.allocateMemory((long)WindowsAsynchronousSocketChannelImpl.SIZEOF_WSABUFARRAY);
        this.writeBufferArray = WindowsAsynchronousSocketChannelImpl.unsafe.allocateMemory((long)WindowsAsynchronousSocketChannelImpl.SIZEOF_WSABUFARRAY);
    }
    
    WindowsAsynchronousSocketChannelImpl(Iocp iocp) throws IOException {
        this(iocp, true);
    }
    
    @Override
    public AsynchronousChannelGroupImpl group() {
        return this.iocp;
    }
    
    @Override
    public <V, A> PendingFuture<V, A> getByOverlapped(long n) {
        return this.ioCache.remove(n);
    }
    
    long handle() {
        return this.handle;
    }
    
    void setConnected(InetSocketAddress localAddress, InetSocketAddress remoteAddress) {
        synchronized (this.stateLock) {
            this.state = 2;
            this.localAddress = localAddress;
            this.remoteAddress = remoteAddress;
        }
    }
    
    @Override
    void implClose() throws IOException {
        closesocket0(this.handle);
        this.ioCache.close();
        WindowsAsynchronousSocketChannelImpl.unsafe.freeMemory(this.readBufferArray);
        WindowsAsynchronousSocketChannelImpl.unsafe.freeMemory(this.writeBufferArray);
        if (this.completionKey != 0) {
            this.iocp.disassociate(this.completionKey);
        }
    }
    
    @Override
    public void onCancel(PendingFuture<?, ?> pendingFuture) {
        if ((pendingFuture.getContext() instanceof WindowsAsynchronousSocketChannelImpl.ConnectTask)) {
            this.killConnect();
        }
        if ((pendingFuture.getContext() instanceof WindowsAsynchronousSocketChannelImpl.ReadTask)) {
            this.killReading();
        }
        if ((pendingFuture.getContext() instanceof WindowsAsynchronousSocketChannelImpl.WriteTask)) {
            this.killWriting();
        }
    }
    
    private void doPrivilegedBind(final SocketAddress socketAddress) throws IOException {
        try {
            AccessController.doPrivileged(new PrivilegedExceptionAction<Void>() {
                @Override
                public Void run() throws IOException {
                    WindowsAsynchronousSocketChannelImpl.this.bind(socketAddress);
                    return null;
                }
            });
        }
        catch (PrivilegedActionException ex) {
            throw (IOException)ex.getException();
        }
    }
    
    @Override
     <A> Future<Void> implConnect(SocketAddress socketAddress, A a, CompletionHandler<Void, ? super A> completionHandler) {
        if (!this.isOpen()) {
            ClosedChannelException ex4 = new ClosedChannelException();
            if (completionHandler == null) {
                return CompletedFuture.withFailure(ex4);
            }
            Invoker.invoke(this, completionHandler, a, null, ex4);
            return null;
        }
        else {
            InetSocketAddress checkAddress = Net.checkAddress(socketAddress);
            SecurityManager securityManager = System.getSecurityManager();
            if (securityManager != null) {
                securityManager.checkConnect(checkAddress.getAddress().getHostAddress(), checkAddress.getPort());
            }
            IOException ex = null;
            synchronized (this.stateLock) {
                if (this.state == 2) {
                    throw new AlreadyConnectedException();
                }
                if (this.state == 1) {
                    throw new ConnectionPendingException();
                }
                if (this.localAddress == null) {
                    try {
                        InetSocketAddress inetSocketAddress = new InetSocketAddress(0);
                        if (securityManager == null) {
                            this.bind(inetSocketAddress);
                        }
                        else {
                            this.doPrivilegedBind(inetSocketAddress);
                        }
                    }
                    catch (IOException ex2) {
                        ex = ex2;
                    }
                }
                if (ex == null) {
                    this.state = 1;
                }
            }
            if (ex != null) {
                try {
                    this.close();
                }
                catch (IOException ex3) {}
                if (completionHandler == null) {
                    return CompletedFuture.withFailure(ex);
                }
                Invoker.invoke(this, completionHandler, a, null, ex);
                return null;
            }
            else {
                PendingFuture o = new PendingFuture(this, completionHandler, a);
                WindowsAsynchronousSocketChannelImpl.ConnectTask context = new WindowsAsynchronousSocketChannelImpl.ConnectTask(checkAddress, o);
                o.setContext(context);
                if (Iocp.supportsThreadAgnosticIo()) {
                    context.run();
                    return o;
                }
                Invoker.invokeOnThreadInThreadPool(this, context);
                return o;
            }
        }
    }
    
    @Override
     <V extends Number, A> Future<V> implRead(boolean b, ByteBuffer byteBuffer, ByteBuffer[] array, long n, TimeUnit timeUnit, A a, CompletionHandler<V, ? super A> completionHandler) {
        PendingFuture pendingFuture = new PendingFuture(this, completionHandler, a);
        if (b) {
        }
        else {
            array = new ByteBuffer[] { byteBuffer };
        }
        final Object context = new WindowsAsynchronousSocketChannelImpl.ReadTask(array, b, pendingFuture);
        pendingFuture.setContext(context);
        if (n > 0L) {
            pendingFuture.setTimeoutTask(this.iocp.schedule(new Runnable() {
                @Override
                public void run() {
                    context.timeout();
                }
            }, n, timeUnit));
        }
        if (Iocp.supportsThreadAgnosticIo()) {
            context.run();
            return pendingFuture;
        }
        Invoker.invokeOnThreadInThreadPool(this, context);
        return pendingFuture;
    }
    
    @Override
     <V extends Number, A> Future<V> implWrite(boolean b, ByteBuffer byteBuffer, ByteBuffer[] array, long n, TimeUnit timeUnit, A a, CompletionHandler<V, ? super A> completionHandler) {
        PendingFuture pendingFuture = new PendingFuture(this, completionHandler, a);
        if (b) {
        }
        else {
            array = new ByteBuffer[] { byteBuffer };
        }
        final Object context = new WindowsAsynchronousSocketChannelImpl.WriteTask(array, b, pendingFuture);
        pendingFuture.setContext(context);
        if (n > 0L) {
            pendingFuture.setTimeoutTask(this.iocp.schedule(new Runnable() {
                @Override
                public void run() {
                    context.timeout();
                }
            }, n, timeUnit));
        }
        if (Iocp.supportsThreadAgnosticIo()) {
            context.run();
            return pendingFuture;
        }
        Invoker.invokeOnThreadInThreadPool(this, context);
        return pendingFuture;
    }
    
    private static native void initIDs();
    
    private static native int connect0(long p0, boolean p1, InetAddress p2, int p3, long p4) throws IOException;
    
    private static native void updateConnectContext(long p0) throws IOException;
    
    private static native int read0(long p0, int p1, long p2, long p3) throws IOException;
    
    private static native int write0(long p0, int p1, long p2, long p3) throws IOException;
    
    private static native void shutdown0(long p0, int p1) throws IOException;
    
    private static native void closesocket0(long p0) throws IOException;
    
    static void access$000(long n) throws IOException {
        updateConnectContext(n);
    }
    
    static int access$200(long n, boolean b, InetAddress inetAddress, int n2, long n3) throws IOException {
        return connect0(n, b, inetAddress, n2, n3);
    }
    
    static int access$700(long n, int n2, long n3, long n4) throws IOException {
        return read0(n, n2, n3, n4);
    }
    
    static int access$900(long n, int n2, long n3, long n4) throws IOException {
        return write0(n, n2, n3, n4);
    }
    
    static {
        unsafe = Unsafe.getUnsafe();
        WindowsAsynchronousSocketChannelImpl.addressSize = WindowsAsynchronousSocketChannelImpl.unsafe.addressSize();
        SIZEOF_WSABUF = dependsArch(8, 16);
        OFFSETOF_BUF = dependsArch(4, 8);
        SIZEOF_WSABUFARRAY = 16 * WindowsAsynchronousSocketChannelImpl.SIZEOF_WSABUF;
        IOUtil.load();
        initIDs();
    }
    
    private class ConnectTask<A> implements Runnable, Iocp.ResultHandler
    {
        private final InetSocketAddress remote;
        private final PendingFuture<Void, A> result;
        
        ConnectTask(InetSocketAddress remote, PendingFuture<Void, A> result) {
            this.remote = remote;
            this.result = result;
        }
        
        private void closeChannel() {
            try {
                WindowsAsynchronousSocketChannelImpl.this.close();
            }
            catch (IOException ex) {}
        }
        
        private IOException toIOException(Throwable cause) {
            if ((cause instanceof IOException)) {
                if ((cause instanceof ClosedChannelException)) {
                    cause = new AsynchronousCloseException();
                }
                return (IOException)cause;
            }
            else {
                return new IOException(cause);
            }
        }
        
        private void afterConnect() throws IOException {
            WindowsAsynchronousSocketChannelImpl.access$000(WindowsAsynchronousSocketChannelImpl.this.handle);
            synchronized (WindowsAsynchronousSocketChannelImpl.this.stateLock) {
                WindowsAsynchronousSocketChannelImpl.this.state = 2;
                WindowsAsynchronousSocketChannelImpl.this.remoteAddress = this.remote;
            }
        }
        
        @Override
        public void run() {
            long add = 0L;
            Throwable t = null;
            try {
                WindowsAsynchronousSocketChannelImpl.this.begin();
                synchronized (this.result) {
                    add = WindowsAsynchronousSocketChannelImpl.this.ioCache.add(this.result);
                    if (WindowsAsynchronousSocketChannelImpl.access$200(WindowsAsynchronousSocketChannelImpl.this.handle, Net.isIPv6Available(), this.remote.getAddress(), this.remote.getPort(), add) == -2) {
                        return;
                    }
                    this.afterConnect();
                    this.result.setResult(null);
                }
            }
            catch (Throwable t2) {
                if (add != 0L) {
                    WindowsAsynchronousSocketChannelImpl.this.ioCache.remove(add);
                }
                t = t2;
            }
            finally {
                WindowsAsynchronousSocketChannelImpl.this.end();
            }
            if (t != null) {
                this.closeChannel();
                this.result.setFailure(this.toIOException(t));
            }
            Invoker.invoke(this.result);
        }
        
        @Override
        public void completed(int n, boolean b) {
            Throwable t = null;
            try {
                WindowsAsynchronousSocketChannelImpl.this.begin();
                this.afterConnect();
                this.result.setResult(null);
            }
            catch (Throwable t2) {
                t = t2;
            }
            finally {
                WindowsAsynchronousSocketChannelImpl.this.end();
            }
            if (t != null) {
                this.closeChannel();
                this.result.setFailure(this.toIOException(t));
            }
            if (b) {
                Invoker.invokeUnchecked(this.result);
            }
            else {
                Invoker.invoke(this.result);
            }
        }
        
        @Override
        public void failed(int n, IOException failure) {
            if (WindowsAsynchronousSocketChannelImpl.this.isOpen()) {
                this.closeChannel();
                this.result.setFailure(failure);
            }
            else {
                this.result.setFailure(new AsynchronousCloseException());
            }
            Invoker.invoke(this.result);
        }
    }
    
    private class ReadTask<V, A> implements Runnable, Iocp.ResultHandler
    {
        private final ByteBuffer[] bufs;
        private final int numBufs;
        private final boolean scatteringRead;
        private final PendingFuture<V, A> result;
        private ByteBuffer[] shadow;
        
        ReadTask(ByteBuffer[] bufs, boolean scatteringRead, PendingFuture<V, A> result) {
            this.bufs = bufs;
            this.numBufs = ((bufs.length > 16) ? 16 : bufs.length);
            this.scatteringRead = scatteringRead;
            this.result = result;
        }
        
        void prepareBuffers() {
            this.shadow = new ByteBuffer[this.numBufs];
            long access$300 = WindowsAsynchronousSocketChannelImpl.this.readBufferArray;
            int i = 0;
            while (i < this.numBufs) {
                ByteBuffer byteBuffer = this.bufs[i];
                int position = byteBuffer.position();
                int limit = byteBuffer.limit();
                assert position <= limit;
                int x = (position <= limit) ? (limit - position) : 0;
                long address;
                if (!(byteBuffer instanceof DirectBuffer)) {
                    ByteBuffer temporaryDirectBuffer = Util.getTemporaryDirectBuffer(x);
                    this.shadow[i] = temporaryDirectBuffer;
                    address = ((DirectBuffer)temporaryDirectBuffer).address();
                }
                else {
                    this.shadow[i] = byteBuffer;
                    address = ((DirectBuffer)byteBuffer).address() + (long)position;
                }
                WindowsAsynchronousSocketChannelImpl.unsafe.putAddress(access$300 + (long)WindowsAsynchronousSocketChannelImpl.OFFSETOF_BUF, address);
                WindowsAsynchronousSocketChannelImpl.unsafe.putInt(access$300 + 0L, x);
                access$300 += (long)WindowsAsynchronousSocketChannelImpl.SIZEOF_WSABUF;
                ++i;
            }
        }
        
        void updateBuffers(int n) {
            int i = 0;
            while (i < this.numBufs) {
                ByteBuffer byteBuffer = this.shadow[i];
                int position = byteBuffer.position();
                int remaining = byteBuffer.remaining();
                if (n >= remaining) {
                    n -= remaining;
                    int n3 = position + remaining;
                    try {
                        byteBuffer.position(n3);
                    }
                    catch (IllegalArgumentException ex2) {}
                    ++i;
                }
                else if (n > 0) {
                    assert (long)(position + n) < 2147483647L;
                    int n2 = position + n;
                    try {
                        byteBuffer.position(n2);
                    }
                    catch (IllegalArgumentException ex) {}
                    break;
                }
                else {
                    break;
                }
            }
            for (int j = 0; j < this.numBufs; ++j) {
                if (!(this.bufs[j] instanceof DirectBuffer)) {
                    this.shadow[j].flip();
                    try {
                        this.bufs[j].put(this.shadow[j]);
                    }
                    catch (BufferOverflowException ex3) {}
                }
            }
        }
        
        void releaseBuffers() {
            for (int i = 0; i < this.numBufs; ++i) {
                if (!(this.bufs[i] instanceof DirectBuffer)) {
                    Util.releaseTemporaryDirectBuffer(this.shadow[i]);
                }
            }
        }
        
        @Override
        public void run() {
            long add = 0L;
            boolean n = false;
            boolean n2 = false;
            try {
                WindowsAsynchronousSocketChannelImpl.this.begin();
                this.prepareBuffers();
                n = true;
                add = WindowsAsynchronousSocketChannelImpl.this.ioCache.add(this.result);
                int access$700 = WindowsAsynchronousSocketChannelImpl.access$700(WindowsAsynchronousSocketChannelImpl.this.handle, this.numBufs, WindowsAsynchronousSocketChannelImpl.this.readBufferArray, add);
                if (access$700 == -2) {
                    n2 = true;
                    return;
                }
                if (access$700 == -1) {
                    WindowsAsynchronousSocketChannelImpl.this.enableReading();
                    if (this.scatteringRead) {
                        this.result.setResult(Long.valueOf(-1L));
                    }
                    else {
                        this.result.setResult(Integer.valueOf(-1));
                    }
                }
                else {
                    throw new InternalError("Read completed immediately");
                }
            }
            catch (Throwable o) {
                WindowsAsynchronousSocketChannelImpl.this.enableReading();
                if ((o instanceof ClosedChannelException)) {
                    o = new AsynchronousCloseException();
                }
                if (!(o instanceof IOException)) {
                    o = new IOException(o);
                }
                this.result.setFailure(o);
            }
            finally {
                if (n2 == 0) {
                    if (add != 0L) {
                        WindowsAsynchronousSocketChannelImpl.this.ioCache.remove(add);
                    }
                    if (!(n == 0)) {
                        this.releaseBuffers();
                    }
                }
                WindowsAsynchronousSocketChannelImpl.this.end();
            }
            Invoker.invoke(this.result);
        }
        
        @Override
        public void completed(int i, boolean b) {
            if (i == 0) {
                i = -1;
            }
            else {
                this.updateBuffers(i);
            }
            this.releaseBuffers();
            synchronized (this.result) {
                if (this.result.isDone()) {
                    return;
                }
                WindowsAsynchronousSocketChannelImpl.this.enableReading();
                if (this.scatteringRead) {
                    this.result.setResult(Long.valueOf((long)i));
                }
                else {
                    this.result.setResult(Integer.valueOf(i));
                }
            }
            if (b) {
                Invoker.invokeUnchecked(this.result);
            }
            else {
                Invoker.invoke(this.result);
            }
        }
        
        @Override
        public void failed(int n, IOException failure) {
            this.releaseBuffers();
            if (!WindowsAsynchronousSocketChannelImpl.this.isOpen()) {
                failure = new AsynchronousCloseException();
            }
            synchronized (this.result) {
                if (this.result.isDone()) {
                    return;
                }
                WindowsAsynchronousSocketChannelImpl.this.enableReading();
                this.result.setFailure(failure);
            }
            Invoker.invoke(this.result);
        }
        
        void timeout() {
            synchronized (this.result) {
                if (this.result.isDone()) {
                    return;
                }
                WindowsAsynchronousSocketChannelImpl.this.enableReading(true);
                this.result.setFailure(new InterruptedByTimeoutException());
            }
            Invoker.invoke(this.result);
        }
    }
    
    private class WriteTask<V, A> implements Runnable, Iocp.ResultHandler
    {
        private final ByteBuffer[] bufs;
        private final int numBufs;
        private final boolean gatheringWrite;
        private final PendingFuture<V, A> result;
        private ByteBuffer[] shadow;
        
        WriteTask(ByteBuffer[] bufs, boolean gatheringWrite, PendingFuture<V, A> result) {
            this.bufs = bufs;
            this.numBufs = ((bufs.length > 16) ? 16 : bufs.length);
            this.gatheringWrite = gatheringWrite;
            this.result = result;
        }
        
        void prepareBuffers() {
            this.shadow = new ByteBuffer[this.numBufs];
            long access$800 = WindowsAsynchronousSocketChannelImpl.this.writeBufferArray;
            int i = 0;
            while (i < this.numBufs) {
                ByteBuffer src = this.bufs[i];
                int position = src.position();
                int limit = src.limit();
                assert position <= limit;
                int x = (position <= limit) ? (limit - position) : 0;
                long address;
                if (!(src instanceof DirectBuffer)) {
                    ByteBuffer temporaryDirectBuffer = Util.getTemporaryDirectBuffer(x);
                    temporaryDirectBuffer.put(src);
                    temporaryDirectBuffer.flip();
                    src.position(position);
                    this.shadow[i] = temporaryDirectBuffer;
                    address = ((DirectBuffer)temporaryDirectBuffer).address();
                }
                else {
                    this.shadow[i] = src;
                    address = ((DirectBuffer)src).address() + (long)position;
                }
                WindowsAsynchronousSocketChannelImpl.unsafe.putAddress(access$800 + (long)WindowsAsynchronousSocketChannelImpl.OFFSETOF_BUF, address);
                WindowsAsynchronousSocketChannelImpl.unsafe.putInt(access$800 + 0L, x);
                access$800 += (long)WindowsAsynchronousSocketChannelImpl.SIZEOF_WSABUF;
                ++i;
            }
        }
        
        void updateBuffers(int n) {
            int i = 0;
            while (i < this.numBufs) {
                ByteBuffer byteBuffer = this.bufs[i];
                int position = byteBuffer.position();
                int limit = byteBuffer.limit();
                limit = ((position <= limit) ? (limit - position) : limit);
                if (n >= limit) {
                    n -= limit;
                    int n3 = position + limit;
                    try {
                        byteBuffer.position(n3);
                    }
                    catch (IllegalArgumentException ex2) {}
                    ++i;
                }
                else if (n > 0) {
                    assert (long)(position + n) < 2147483647L;
                    int n2 = position + n;
                    try {
                        byteBuffer.position(n2);
                    }
                    catch (IllegalArgumentException ex) {}
                    break;
                }
                else {
                    break;
                }
            }
        }
        
        void releaseBuffers() {
            for (int i = 0; i < this.numBufs; ++i) {
                if (!(this.bufs[i] instanceof DirectBuffer)) {
                    Util.releaseTemporaryDirectBuffer(this.shadow[i]);
                }
            }
        }
        
        @Override
        public void run() {
            long add = 0L;
            boolean n = false;
            boolean n2 = false;
            boolean n3 = false;
            try {
                WindowsAsynchronousSocketChannelImpl.this.begin();
                this.prepareBuffers();
                n = true;
                add = WindowsAsynchronousSocketChannelImpl.this.ioCache.add(this.result);
                int access$900 = WindowsAsynchronousSocketChannelImpl.access$900(WindowsAsynchronousSocketChannelImpl.this.handle, this.numBufs, WindowsAsynchronousSocketChannelImpl.this.writeBufferArray, add);
                if (access$900 == -2) {
                    n2 = true;
                    return;
                }
                if (access$900 == -1) {
                    n3 = true;
                    throw new ClosedChannelException();
                }
                throw new InternalError("Write completed immediately");
            }
            catch (Throwable o) {
                WindowsAsynchronousSocketChannelImpl.this.enableWriting();
                if (n3 == 0 && o instanceof ClosedChannelException) {
                    o = new AsynchronousCloseException();
                }
                if (!(o instanceof IOException)) {
                    o = new IOException(o);
                }
                this.result.setFailure(o);
            }
            finally {
                if (n2 == 0) {
                    if (add != 0L) {
                        WindowsAsynchronousSocketChannelImpl.this.ioCache.remove(add);
                    }
                    if (!(n == 0)) {
                        this.releaseBuffers();
                    }
                }
                WindowsAsynchronousSocketChannelImpl.this.end();
            }
            Invoker.invoke(this.result);
        }
        
        @Override
        public void completed(int i, boolean b) {
            this.updateBuffers(i);
            this.releaseBuffers();
            synchronized (this.result) {
                if (this.result.isDone()) {
                    return;
                }
                WindowsAsynchronousSocketChannelImpl.this.enableWriting();
                if (this.gatheringWrite) {
                    this.result.setResult(Long.valueOf((long)i));
                }
                else {
                    this.result.setResult(Integer.valueOf(i));
                }
            }
            if (b) {
                Invoker.invokeUnchecked(this.result);
            }
            else {
                Invoker.invoke(this.result);
            }
        }
        
        @Override
        public void failed(int n, IOException failure) {
            this.releaseBuffers();
            if (!WindowsAsynchronousSocketChannelImpl.this.isOpen()) {
                failure = new AsynchronousCloseException();
            }
            synchronized (this.result) {
                if (this.result.isDone()) {
                    return;
                }
                WindowsAsynchronousSocketChannelImpl.this.enableWriting();
                this.result.setFailure(failure);
            }
            Invoker.invoke(this.result);
        }
        
        void timeout() {
            synchronized (this.result) {
                if (this.result.isDone()) {
                    return;
                }
                WindowsAsynchronousSocketChannelImpl.this.enableWriting(true);
                this.result.setFailure(new InterruptedByTimeoutException());
            }
            Invoker.invoke(this.result);
        }
    }
}
